
#include "funct.cuh"


void nlp_pnt_jac_create_const_arr(NLP_data* nlp_data)
{
	int lint[1];

	lint[0] = nlp_offset_to_warp(nlp_data->comb_pnt_n) * warp_size;

	cudaMalloc(&nlp_data->d_nlp_pnt_jac_base_comb, sizeof(double) * lint[0]);

	nlp_data->nlp_pnt_sol->h_nlp_pnt_jac_exp_comb_int_data =
		new int[nlp_exp_comb_int_data_length];
	nlp_data->nlp_pnt_sol->h_nlp_pnt_jac_exp_comb_dbl_ptr =
		new double*[nlp_exp_comb_dbl_ptr_length];

	cudaMalloc(&nlp_data->nlp_pnt_sol->d_nlp_pnt_jac_exp_comb_int_data,
		sizeof(int) * nlp_exp_comb_int_data_length);
	cudaMalloc(&nlp_data->nlp_pnt_sol->d_nlp_pnt_jac_exp_comb_dbl_ptr,
		sizeof(double*) * nlp_exp_comb_dbl_ptr_length);

	nlp_data->nlp_pnt_jac_arr_created = 1;
}


void nlp_pnt_jacobian(NLP_data* nlp_data, Algo_dat* algo_dat, Timer_dat* timer_dat)
{
	NLP_pnt_jac nlp_pnt_jac;

	cudaStream_t* s;
	cudaEvent_t* e;

	nlp_pnt_jac_create_streams(s, e);

	nlp_pnt_jac_pre_setup(&nlp_pnt_jac, nlp_data);

	nlp_pnt_init_arr(&nlp_pnt_jac);

	nlp_pnt_jac_post_init_arr_setup(&nlp_pnt_jac, nlp_data,
		s[0], e[nlp_pnt_jac_stream_cnt]);

	cudaEventRecord(e[nlp_pnt_jac_stream_cnt], s[0]);
}


void nlp_pnt_jac_create_streams(cudaStream_t*& s, cudaEvent_t*& e)
{
	int lint1;

	s = new cudaStream_t[nlp_pnt_jac_stream_cnt];
	e = new cudaEvent_t[nlp_pnt_jac_stream_cnt + 2];

	for (lint1 = 0; lint1 < nlp_pnt_jac_stream_cnt; lint1++)
	{
		cudaStreamCreate(&s[lint1]);
	}

	for (lint1 = 0; lint1 < (nlp_pnt_jac_stream_cnt + 2); lint1++)
	{
		cudaEventCreate(&e[lint1]);
	}
}


void nlp_pnt_jac_pre_setup(NLP_pnt_jac* nlp_pnt_jac, NLP_data* nlp_data)
{
	nlp_pnt_jac->coeff_cnt = nlp_data->comb_pnt_n;
	nlp_pnt_jac->coeff_cnt_plus1 = nlp_data->comb_pnt_n + 1;
	nlp_pnt_jac->comb_pnt_n = nlp_data->comb_pnt_n;
	nlp_pnt_jac->issues_n = nlp_data->algo_dat->issues_n;

	nlp_pnt_jac->no_prev_mul_sum = 1;
	nlp_pnt_jac->found_better_case = 0;

	nlp_pnt_jac->lmt_iss_res = nlp_pnt_jac_lmt_iss_res;

	nlp_pnt_jac->coeff_incr = nlp_data->res;

	nlp_pnt_jac->d_base_coeff = nlp_data->d_nlp_pnt_jac_base_comb;

	if ((nlp_pnt_jac->issues_n * 2) > nlp_matrix_mul_transpose_block_size)
	{
		nlp_pnt_jac->mul_transpose_cases_per_block = 1;
		nlp_pnt_jac->use_mul_transpose_big = 1;
	}

	else
	{
		nlp_pnt_jac->mul_transpose_cases_per_block =
			nlp_matrix_mul_transpose_block_size / nlp_pnt_jac->issues_n;
		nlp_pnt_jac->use_mul_transpose_big = 0;
	}

	if (nlp_pnt_jac->coeff_cnt_plus1 < linear_eq_lmt)
	{
		nlp_pnt_jac->use_le_big = 0;
	}

	else
	{
		nlp_pnt_jac->use_le_big = 1;
	}
}


void nlp_pnt_init_arr(NLP_pnt_jac* nlp_pnt_jac)
{
	int lint[5];

	nlp_pnt_jac->cases_per_krnl_offset =
		nlp_offset_to_warp(nlp_pnt_jac_cases_per_krnl) * warp_size;

	nlp_pnt_jac->coeff_offset =
		nlp_offset_to_warp(nlp_pnt_jac->coeff_cnt) * warp_size;

	nlp_pnt_jac->iss_offset =
		nlp_offset_to_warp(nlp_pnt_jac->issues_n) * warp_size;

	nlp_pnt_jac->matrix_offset =
		nlp_offset_to_warp(nlp_pnt_jac->coeff_cnt_plus1) * warp_size;

	nlp_pnt_jac->nlp_distr_and_adj_comb_pnt_blockdim =
		nlp_offset_to_x(nlp_pnt_jac->coeff_cnt, nlp_distr_and_adj_comb_block_size);

	nlp_pnt_jac->nlp_distr_and_adj_comb_int_data_offset =
		nlp_offset_to_warp(nlp_distr_and_adj_comb_int_data_length) * warp_size;

	nlp_pnt_jac->nlp_distr_and_adj_comb_dbl_data_offset =
		nlp_offset_to_warp(nlp_distr_and_adj_comb_dbl_data_length) * warp_size;

	nlp_pnt_jac->nlp_distr_and_adj_comb_dbl_ptr_offset =
		nlp_offset_to_warp(nlp_distr_and_adj_comb_dbl_ptr_length) * warp_size;

	nlp_pnt_jac->nlp_exp_comb_int_data_offset =
		nlp_offset_to_warp(nlp_exp_comb_int_data_length) * warp_size;

	nlp_pnt_jac->nlp_exp_comb_dbl_ptr_offset =
		nlp_offset_to_warp(nlp_exp_comb_dbl_ptr_length) * warp_size;

	nlp_pnt_jac->nlp_mul_comb_iss_res_int_data_offset =
		nlp_offset_to_warp(nlp_mul_comb_iss_res_int_data_length) * warp_size;

	nlp_pnt_jac->nlp_mul_comb_iss_res_int_ptr_offset =
		nlp_offset_to_warp(nlp_mul_comb_iss_res_int_ptr_length) * warp_size;

	nlp_pnt_jac->nlp_mul_comb_iss_res_dbl_ptr_offset =
		nlp_offset_to_warp(nlp_mul_comb_iss_res_dbl_ptr_length) * warp_size;

	nlp_pnt_jac->nlp_iss_var_mul_iss_res_int_data_offset =
		nlp_offset_to_warp(nlp_iss_var_mul_iss_res_int_data_length) * warp_size;

	nlp_pnt_jac->nlp_iss_var_mul_iss_res_dbl_ptr_offset =
		nlp_offset_to_warp(nlp_iss_var_mul_iss_res_dbl_ptr_length) * warp_size;

	nlp_pnt_jac->h_nlp_distr_and_adj_comb_int_data =
		new int[nlp_pnt_jac->nlp_distr_and_adj_comb_int_data_offset * nlp_pnt_jac_stream_cnt];
	nlp_pnt_jac->h_nlp_distr_and_adj_comb_dbl_data =
		new double[nlp_pnt_jac->nlp_distr_and_adj_comb_dbl_data_offset * nlp_pnt_jac_stream_cnt];
	nlp_pnt_jac->h_nlp_distr_and_adj_comb_dbl_ptr =
		new double*[nlp_pnt_jac->nlp_distr_and_adj_comb_dbl_ptr_offset * nlp_pnt_jac_stream_cnt];

	cudaMalloc(&nlp_pnt_jac->d_nlp_distr_and_adj_comb_int_data,
		sizeof(int) * nlp_pnt_jac->nlp_distr_and_adj_comb_int_data_offset * nlp_pnt_jac_stream_cnt);
	cudaMalloc(&nlp_pnt_jac->d_nlp_distr_and_adj_comb_dbl_data,
		sizeof(double) * nlp_pnt_jac->nlp_distr_and_adj_comb_dbl_data_offset * nlp_pnt_jac_stream_cnt);
	cudaMalloc(&nlp_pnt_jac->d_nlp_distr_and_adj_comb_dbl_ptr,
		sizeof(double*) * nlp_pnt_jac->nlp_distr_and_adj_comb_dbl_ptr_offset * nlp_pnt_jac_stream_cnt);

	nlp_pnt_jac->h_nlp_exp_comb_int_data =
		new int[nlp_pnt_jac->nlp_exp_comb_int_data_offset * nlp_pnt_jac_stream_cnt];
	nlp_pnt_jac->h_nlp_exp_comb_dbl_ptr =
		new double*[nlp_pnt_jac->nlp_exp_comb_dbl_ptr_offset * nlp_pnt_jac_stream_cnt];

	cudaMalloc(&nlp_pnt_jac->d_nlp_exp_comb_int_data,
		sizeof(int) * nlp_pnt_jac->nlp_exp_comb_int_data_offset * nlp_pnt_jac_stream_cnt);
	cudaMalloc(&nlp_pnt_jac->d_nlp_exp_comb_dbl_ptr,
		sizeof(double*) * nlp_pnt_jac->nlp_exp_comb_dbl_ptr_offset * nlp_pnt_jac_stream_cnt);

	nlp_pnt_jac->h_nlp_iss_var_lmt_iss_res_int_data =
		new int[nlp_iss_var_lmt_iss_res_int_data_length];
	nlp_pnt_jac->h_nlp_iss_var_lmt_iss_res_dbl_data =
		new double[nlp_iss_var_lmt_iss_res_dbl_data_length];
	nlp_pnt_jac->h_nlp_iss_var_lmt_iss_res_dbl_ptr =
		new double*[nlp_iss_var_lmt_iss_res_dbl_ptr_length];

	cudaMalloc(&nlp_pnt_jac->d_nlp_iss_var_lmt_iss_res_int_data,
		sizeof(int) * nlp_iss_var_lmt_iss_res_int_data_length);
	cudaMalloc(&nlp_pnt_jac->d_nlp_iss_var_lmt_iss_res_dbl_data,
		sizeof(double) * nlp_iss_var_lmt_iss_res_dbl_data_length);
	cudaMalloc(&nlp_pnt_jac->d_nlp_iss_var_lmt_iss_res_dbl_ptr,
		sizeof(double*) * nlp_iss_var_lmt_iss_res_dbl_ptr_length);

	nlp_pnt_jac->h_nlp_mul_comb_iss_res_int_data =
		new int[nlp_pnt_jac->nlp_mul_comb_iss_res_int_data_offset * nlp_pnt_jac_stream_cnt];
	nlp_pnt_jac->h_nlp_mul_comb_iss_res_int_ptr =
		new int*[nlp_pnt_jac->nlp_mul_comb_iss_res_int_ptr_offset * nlp_pnt_jac_stream_cnt];
	nlp_pnt_jac->h_nlp_mul_comb_iss_res_dbl_ptr =
		new double*[nlp_pnt_jac->nlp_mul_comb_iss_res_dbl_ptr_offset * nlp_pnt_jac_stream_cnt];

	cudaMalloc(&nlp_pnt_jac->d_nlp_mul_comb_iss_res_int_data,
		sizeof(int) * nlp_pnt_jac->nlp_mul_comb_iss_res_int_data_offset * nlp_pnt_jac_stream_cnt);
	cudaMalloc(&nlp_pnt_jac->d_nlp_mul_comb_iss_res_int_ptr,
		sizeof(int*) * nlp_pnt_jac->nlp_mul_comb_iss_res_int_ptr_offset * nlp_pnt_jac_stream_cnt);
	cudaMalloc(&nlp_pnt_jac->d_nlp_mul_comb_iss_res_dbl_ptr,
		sizeof(double*) * nlp_pnt_jac->nlp_mul_comb_iss_res_dbl_ptr_offset * nlp_pnt_jac_stream_cnt);

	nlp_pnt_jac->h_nlp_iss_var_mul_iss_res_int_data =
		new int[nlp_pnt_jac->nlp_iss_var_mul_iss_res_int_data_offset * nlp_pnt_jac_stream_cnt];
	nlp_pnt_jac->h_nlp_iss_var_mul_iss_res_dbl_ptr =
		new double*[nlp_pnt_jac->nlp_iss_var_mul_iss_res_dbl_ptr_offset * nlp_pnt_jac_stream_cnt];

	cudaMalloc(&nlp_pnt_jac->d_nlp_iss_var_mul_iss_res_int_data,
		sizeof(int) * nlp_pnt_jac->nlp_iss_var_mul_iss_res_int_data_offset * nlp_pnt_jac_stream_cnt);
	cudaMalloc(&nlp_pnt_jac->d_nlp_iss_var_mul_iss_res_dbl_ptr,
		sizeof(double*) * nlp_pnt_jac->nlp_iss_var_mul_iss_res_dbl_ptr_offset * nlp_pnt_jac_stream_cnt);

	nlp_pnt_jac->h_nlp_matrix_mul_transpose_int_data =
		new int[nlp_matrix_mul_transpose_int_data_length];
	nlp_pnt_jac->h_nlp_matrix_mul_transpose_dbl_ptr =
		new double*[nlp_matrix_mul_transpose_dbl_ptr_length];

	cudaMalloc(&nlp_pnt_jac->d_nlp_matrix_mul_transpose_int_data,
		sizeof(int) * nlp_matrix_mul_transpose_int_data_length);
	cudaMalloc(&nlp_pnt_jac->d_nlp_matrix_mul_transpose_dbl_ptr,
		sizeof(double*) * nlp_matrix_mul_transpose_dbl_ptr_length);

	nlp_pnt_jac->h_nlp_matrix_mul_transpose_big_int_data =
		new int[nlp_matrix_mul_transpose_big_int_data_length];
	nlp_pnt_jac->h_nlp_matrix_mul_transpose_big_dbl_ptr =
		new double*[nlp_matrix_mul_transpose_big_dbl_ptr_length];

	cudaMalloc(&nlp_pnt_jac->d_nlp_matrix_mul_transpose_big_int_data,
		sizeof(int) * nlp_matrix_mul_transpose_big_int_data_length);
	cudaMalloc(&nlp_pnt_jac->d_nlp_matrix_mul_transpose_big_dbl_ptr,
		sizeof(double*) * nlp_matrix_mul_transpose_big_dbl_ptr_length);

	nlp_pnt_jac->h_nlp_linear_eq_int_data =
		new int[nlp_linear_eq_int_data_length];
	nlp_pnt_jac->h_nlp_linear_eq_dbl_ptr =
		new double*[nlp_linear_eq_dbl_ptr_length];

	cudaMalloc(&nlp_pnt_jac->d_nlp_linear_eq_int_data,
		sizeof(int) * nlp_linear_eq_int_data_length);
	cudaMalloc(&nlp_pnt_jac->d_nlp_linear_eq_dbl_ptr,
		sizeof(double*) * nlp_linear_eq_dbl_ptr_length);

	nlp_pnt_jac->h_nlp_linear_eq_big_int_data =
		new int[nlp_linear_eq_big_int_data_length];
	nlp_pnt_jac->h_nlp_linear_eq_big_dbl_ptr =
		new double*[nlp_linear_eq_big_dbl_ptr_length];

	cudaMalloc(&nlp_pnt_jac->d_nlp_linear_eq_big_int_data,
		sizeof(int) * nlp_linear_eq_big_int_data_length);
	cudaMalloc(&nlp_pnt_jac->d_nlp_linear_eq_big_dbl_ptr,
		sizeof(double*) * nlp_linear_eq_big_dbl_ptr_length);

	nlp_pnt_jac->h_init_krnl = new int[nlp_pnt_jac_stream_cnt];

	nlp_pnt_jac->h_base_coeff = new double[nlp_pnt_jac->comb_pnt_n];
	cudaMallocHost(&nlp_pnt_jac->h_out_mul_sum,
		sizeof(double) * 3);
	cudaMallocHost(&nlp_pnt_jac->h_le_out_status, sizeof(double));
	cudaMallocHost(&nlp_pnt_jac->h_le_out_arr,
		sizeof(double) * nlp_pnt_jac->coeff_cnt);
	cudaMallocHost(&nlp_pnt_jac->h_mul_sum_store, sizeof(double));

	lint[0] = nlp_pnt_jac->coeff_offset * nlp_pnt_jac_cases_per_krnl *
		nlp_pnt_jac_stream_cnt;
	lint[1] = nlp_pnt_jac->cases_per_krnl_offset * nlp_pnt_jac_stream_cnt;
	lint[2] = nlp_pnt_jac->iss_offset * nlp_pnt_jac_cases_per_krnl *
		nlp_pnt_jac_stream_cnt;
	lint[3] = nlp_pnt_jac->iss_offset * nlp_pnt_jac->coeff_cnt;
	lint[4] = nlp_pnt_jac->matrix_offset * nlp_pnt_jac->coeff_cnt;

	cudaMalloc(&nlp_pnt_jac->d_work_comb,
		sizeof(double) * lint[0]);
	cudaMalloc(&nlp_pnt_jac->d_base_iss_mul_sum,
		sizeof(double) * nlp_pnt_jac->issues_n);
	cudaMalloc(&nlp_pnt_jac->d_out_mul_sum,
		sizeof(double) * lint[1]);
	cudaMalloc(&nlp_pnt_jac->d_out_iss_mul_sum,
		sizeof(double) * lint[2]);
	cudaMalloc(&nlp_pnt_jac->d_base_iss_res,
		sizeof(double) * nlp_pnt_jac->issues_n);
	cudaMalloc(&nlp_pnt_jac->d_out_iss_res,
		sizeof(double) * lint[2]);

	cudaMalloc(&nlp_pnt_jac->d_coeff_delta,
		sizeof(double) * lint[2]);
	cudaMalloc(&nlp_pnt_jac->d_jac_matrix,
		sizeof(double) * lint[3]);
	cudaMalloc(&nlp_pnt_jac->d_le_matrix,
		sizeof(double) * lint[4]);
	cudaMalloc(&nlp_pnt_jac->d_le_out_status, sizeof(double));
	cudaMalloc(&nlp_pnt_jac->d_le_out_arr,
		sizeof(double) * nlp_pnt_jac->coeff_cnt);
	cudaMalloc(&nlp_pnt_jac->d_le_temp_out_arrA,
		sizeof(double) * nlp_pnt_jac->coeff_cnt);
	cudaMalloc(&nlp_pnt_jac->d_le_temp_out_arrB,
		sizeof(double) * nlp_pnt_jac->coeff_cnt);
}


void nlp_pnt_jac_post_init_arr_setup(NLP_pnt_jac* nlp_pnt_jac, NLP_data* nlp_data,
	cudaStream_t s0, cudaEvent_t trigger1)
{
	int lint[0];

	for (lint[0] = 0; lint[0] < nlp_pnt_jac_stream_cnt; lint[0]++)
	{
		nlp_pnt_jac->h_init_krnl[lint[0]] = 1;
	}

	for (lint[0] = 0; lint[0] < nlp_pnt_jac->comb_pnt_n; lint[0]++)
	{
		nlp_pnt_jac->h_base_coeff[lint[0]] = nlp_data->h_port_uni_fwd;
	}

	cudaMemcpyAsync(nlp_pnt_jac->d_base_coeff,
		nlp_pnt_jac->h_base_coeff, sizeof(double) *
		nlp_pnt_jac->comb_pnt_n, cudaMemcpyHostToDevice, s0);

	cudaMemcpyAsync(nlp_pnt_jac->d_coeff_delta,
		&nlp_pnt_jac->coeff_incr, sizeof(double),
		cudaMemcpyHostToDevice, s0);

	cudaEventRecord(trigger1, s0);
}


int nlp_offset_to_warp(int org_size)
{
	int lint1;

	lint1 = org_size / warp_size;

	if ((org_size % warp_size) > 0)
	{
		lint1++;
	}

	return lint1;
}


int nlp_offset_to_x(int org_size, int base)
{
	int lint1;

	lint1 = org_size / base;

	if ((org_size % base) > 0)
	{
		lint1++;
	}

	return lint1;
}
